"""
Fitness assignment processes.

Fitness assignment processes are modules of the fully-configurable
Evolutionary Algorithm :class:`~moptipy.algorithms.so.general_ea.GeneralEA`.
The transform objective values to scalar fitness values that are then used
by :class:`~moptipy.algorithms.modules.selection.Selection` algorithms.
"""
