"""
Numerical optimization algorithms working on real-valued vectors.

The search space here is always
:class:`~moptipy.spaces.vectorspace.VectorSpace`, which prescribes a box in
the space of `n`-dimensional real vectors that is constraint by an upper and a
lower bound in each dimension.
Algorithms working on such spaces can use the rich mathematical relationships
and operations defined over the continuous spaces.
"""
